#!/usr/bin/perl 

#use lib '/home/zhidongx/test/bin';

use strict;
use warnings;
use Getopt::Long;
use Math::Trig;
use ThreaDom;



my $operating_system=$^O;
unless ($operating_system eq "linux"){
    printf("Your operating system $operating_system is unsupported at this time\n");
    printf("Currently only Linux is supported\n");
    exit();
}

my ($libdir,$tlibdir,$seqname,$datadir,$cutoffs);
GetOptions('libdir:s' => \$libdir, 'tlibdir:s'=> \$tlibdir,'seqname:s' => \$seqname, 'datadir:s' => \$datadir,'-cutoffs:s' => \$cutoffs);

if(!$seqname )
{
    printf "ThreaDom USAGE:\n";
    printf "Mandatory arguments:\n";
    printf "./ThreaDom.pl -seqname sequence_name\n";
    printf "Optional arguments whose default values have been set in the config.ini of this package.\n";
    printf "User can also replace the default values in the config file,like:\n";
    printf "      -datadir datadirectory: the Data_path item in the config.ini\n";
    printf "      -tlibdir threadompackagelocation: the  ThreaDom_path item in the config.ini\n";
    printf "      -libdir threadompackagelocation: the I-TASSER_path item in the config.ini\n";
    printf "      -cutoff  \"ce cm ch \"; ce: cutoff value of easy,\tcm :cutoff value of medium;\tch:cutoff value of hard\n";
    
   exit();
}


my $helper       = new DomainHelper();
my %CONFIG       = $helper->readConfig();
my $dataset_name = $CONFIG{'DATASET_NAME'};
my $threadom_path = $CONFIG{'THREADOM_PATH'};
my $ITASSER_path  = $CONFIG{'I-TASSER_PATH'};
my $dssp_path= $CONFIG{'DSSP_PATH'};
my $dp_path= $CONFIG{'DOMAINPARSER_PATH'};
my $pdball_path = $CONFIG{'PDBALL_PATH'};
my $dataset_path  = $CONFIG{'DATA_PATH'};  #where input/output files are
my $cutoff_values = $CONFIG{'CUTOFFS'};

#########set domain parser enviroment  ###############

$ENV{'DSSP_PATH'} = "$dssp_path";
my $domainparser_dir = $dp_path;


my @ss;
if($seqname){
  push( @ss, $seqname);
}
if($datadir){
    $dataset_path=$datadir;
}
if($libdir){
$ITASSER_path=$libdir;
}

if($tlibdir){
$threadom_path=$tlibdir;
}

if(!$pdball_path){
$pdball_path="$ITASSER_path/PDB";
}

if($cutoffs){
$cutoff_values=$cutoffs;
}

my ($ce,$cm,$ch);
$cutoff_values =~ /(\S+)\s+(\S+)\s+(\S+)/;
$ce=$1;
$cm=$2;
$ch=$3;

###cofig the directory carefully for the program######
my $cathdatadir = "$threadom_path/lib";
my $myCathFile  = "$threadom_path/lib/myCath.lib";
my $mapDir      = "$threadom_path/map";
my $dpDBFile    = "$threadom_path/lib/pddb.lib";

#####################################################





my %method2Map;
my %cathDB;
my %myCathDB;
my %dpDB;
my $numDpDB = 0;
my %mapDB;

#---------------------------------------------------------------------------
# set Global variables for INIT.
#---------------------------------------------------------------------------

my $GnumTemplates = 50;
my $num_list      = 0;    #number of list
my %test;                 #store the domain data

my $Lch = 0;              # Length of the sequence

my %res;
my @inits = ();
my $M     = 11;           #### in an order of confidence
my %zscore0;

my $Nt     = 0;           # number of templates
my $type   = "";          # easy medn hard
my $n_good = 0;           # number of good templates
my $M0     = 0;           # zscore is larger than the cutoff

my %L_ali;                # the number of the aligned atoms of the ith template
my %zscore;               # the zscore of the ith template
my %template;             # for example,"3k2zA1"
my %ori;                  # algorithm

my %LINE;

&main();

sub main {
    my ($par1) = @_;
    &initialize();
    &readMyCathDB($myCathFile);
    &initializeMapDB($mapDir);
    &readDpDB($dpDBFile);

    for my $s (@ss) {
        $GnumTemplates = 50;
        &readInit($s);
        &match( $s, $GnumTemplates );
        ###ensure the method match is called.
        &predict( $dataset_path, '', $s, $ce,$cm,$ch);
    }
    &writeDpDB($dpDBFile);
   
}    # ----------  end of subroutine main  ----------

sub readInit {
    my ($s) = @_;
    my $datadir ="$dataset_path";
    open( my $r_init, "$datadir/init.dat" );
    <$r_init> =~ /(\d+)\s+(\S+)\s+(\S+)\s+(\S+)/;
    $Nt     = $1;
    $type   = $2;
    $n_good = $3;
    $M0     = $4;
    undef %L_ali;       # the number of the aligned atoms of the ith template
    undef %zscore;      # the zscore of the ith template
    undef %template;    # for example,"3k2zA1"
    undef %ori;         # algorithm

    for ( my $i = 1 ; $i <= $Nt ; $i++ ) {    # the ith template
        <$r_init> =~ /(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)/;
        $L_ali{$i}  = $1;  # the number of the aligned atoms of the ith template
        $zscore{$i} = $2;  # the zscore of the ith template
        $template{$i} = $4;    # for example,"3k2zA1"
        $ori{$i}      = $5;    # algorithm
        for ( my $j = 1 ; $j <= $L_ali{$i} ; $j++ )
        {                      # read every aligned residue from the file.
            my $line = <$r_init>;
            $LINE{ $i, $j } =
              $line; # the ith template,jth residue is saved in map $LINE{$i,$j}
            $res{ $i, $j } = substr( $line, 22, 4 )
              ;      # The  position of the jth residue of the target sequences
                     # ith template 's jth residue
        }
        <$r_init>;
    }
    close($r_init);
    open( RD, "$datadir/rmsinp" );
    <RD>;
    $Lch = <RD>;
    chop($Lch);
    close(RD);
}    # ----------  end of subroutine readInit  ----------

sub match {
    my ( $s, $numTemplates ) = @_;
    my $ngap            = 0;
    my $nfull           = 0;
    my $nstrong         = 0;
    my $M2              = $M0 / 2;
    my $L_gap           = 80;
    my $aTemplate       = "";
    my $aNormalTemplate = "";
    my %templatesInfo;
    my %origin;
    my %tgap;
    my %II;
    my %ii;
    my %lgap;
    my @arrayHitGap;
    my @arrayDomainSearch_bad;
    my @arrayDPSearch_bad;
    my @arrayDomainSearch_good;
    my @arrayDPSearch_good;
    my @arrayTerminalSearch;
    my @arrayGapSearch;

    my @arrayTemplateInfo;

    $numTemplates=$Nt if($numTemplates>$Nt);
    for ( my $j = 0 ; $j < $Lch ; $j++ ) {
        $arrayDomainSearch_bad[$j]  = 0;
        $arrayDPSearch_bad[$j]      = 0;
        $arrayGapSearch[$j]         = 0;
        $arrayDomainSearch_good[$j] = 0;
        $arrayDPSearch_good[$j]     = 0;
        $arrayTerminalSearch[$j]    = 0;
    }
    
    for ( my $i = 1 ; $i <= $numTemplates ; $i++ ) {
        my @targetNO;
        my @templateNO;

        for ( my $j = 1 ; $j < $L_ali{$i} ; $j++ ) {

            $res{ $i, $j } =~ /(\d+)/;
            $targetNO[ $j - 1 ] = $1;

            my $temp = substr( $LINE{ $i, $j }, 55, 4 );
            $temp =~ /(\d+)/;
            $templateNO[ $j - 1 ] = $1;
        }

        my $z0     = $zscore0{ $ori{$i} };    # the cutoff of zscore.
        my $zscore = $zscore{$i};
        my $arrayDomainSearch;
        my $arrayDPSearch;
        my $isGood = 0;
        if ( $zscore > $z0 ) {
            $arrayDomainSearch = \@arrayDomainSearch_good;
            $arrayDPSearch     = \@arrayDPSearch_good;
            $isGood            = 1;
        }
        else {
            $arrayDomainSearch = \@arrayDomainSearch_bad;
            $arrayDPSearch     = \@arrayDPSearch_bad;
            $isGood            = 0;
        }
        $II{ $ori{$i} }++;
        if ( $ori{$i} eq 'WWW' ) {
            next;
        }
        my $method   = $ori{$i};
        my $aMapType = $method2Map{$method};

        my $templateInfo = "";
        if( not $template{$i}){
           print "warining: no nough templates\n"; 
         last;
         }
        $aNormalTemplate = &normalizeTemplateName("$template{$i}");
        $aTemplate       = $template{$i};
        if ( $aTemplate ne "SingleDomain" ) {
            my %temp = &getMyCathDomainInfo($aNormalTemplate);

            if (%temp) {
                my $alignedRange    = '';
                my $isDiscontinuous = 0;
#                print "$aTemplate:$aNormalTemplate\n";
                for ( my $k = 1 ; $k <= $temp{'numDomains'} ; $k++ ) {
                    my $str1          = "domain$k";
                    my $domains       = $temp{$str1};
                    my @array_d       = split( /\|/, $domains );
                    my $alignedDomain = '';
                    foreach $domains (@array_d) {
                        $domains =~ /(\d+)-(\d+)/;
                        my $nStart1 = $1;
                        my $nEnd    = $2;
                        if ( not defined $nStart1 || not $nEnd ) {
                            print("==$domains==\n");
                        }

                        my $nStart =
                          &searchMapIndex( $aMapType, $aTemplate, $nStart1 );
                        $nEnd = &searchMapIndex( $aMapType, $aTemplate, $nEnd );

                        my $indexStart = -1;
                        my $nMapStart  = -1;
                        my $indexEnd   = -1;
                        my $nMapEnd    = -1;

                        if ( $nStart >= 0 ) {
                            $indexStart =
                              &searchArrayIndex( \@templateNO, $nStart );
                            $nMapStart = $targetNO[$indexStart];
                            $$arrayDomainSearch[$nMapStart] += 1;
                        }
                        if ( $nEnd > 0 ) {
                            $indexEnd =
                              &searchArrayIndex( \@templateNO, $nEnd );
                            $nMapEnd = $targetNO[$indexEnd];
                            $$arrayDomainSearch[$nMapEnd] += 1;
                        }

                        if (    $nMapStart != -1
                            and $nMapEnd != -1
                            and $nMapStart != $nMapEnd )
                        {
                            if ( $alignedDomain ne '' ) {
                                $alignedDomain .= "|$nMapStart-$nMapEnd";
                            }
                            else {
                                $alignedDomain = "$nMapStart-$nMapEnd";
                            }
                        }
                    }
                    if ( $alignedDomain ne '' ) {
                        $alignedRange .= "($alignedDomain)";
                    }

                }
                if ( $alignedRange ne '' ) {
                    $isDiscontinuous = 1 if ( $alignedRange =~ /\|/ );
                    $templateInfo =
"$method\t$aTemplate\t$aNormalTemplate\t$isDiscontinuous\t$isGood\t$alignedRange";
                    push( @arrayTemplateInfo, $templateInfo );
#                    print "$templateInfo\n";
                }

            }
            else {

                my $temp = &getDPDomainInfo($aTemplate);
                for ( my $i = 0 ; $i < $#$temp ; $i++ ) {

                    my $link = $$temp[$i];
                    if ( $link <= $templateNO[$#templateNO] ) {

                        my $indexLink =
                          &searchArrayIndex( \@templateNO, $link );
                        my $nMapLink = $targetNO[$indexLink];
                        $$arrayDPSearch[$nMapLink] += 1;
                    }
                }
            }
        }

        my $indexStart = 0;
        my $nMapStart  = $targetNO[$indexStart];
        $arrayTerminalSearch[$nMapStart] += 1;

        for ( my $i = 1 ; $i <= $#templateNO ; $i++ ) {

            if ( ( $templateNO[$i] - $templateNO[ $i - 1 ] ) > 15 ) {
                my $nGap = $targetNO[ $i - 1 ];
                my $cGap = $targetNO[$i];
                my $med  = int( ( $nGap + $cGap ) / 2 );
                $arrayGapSearch[$nGap] += 1;
            }

        }
        my $indexEnd = $#templateNO;
        my $nMapEnd  = $targetNO[$indexEnd];

        $arrayTerminalSearch[$nMapEnd] += 1;

    }

    #---------------------------------------------------------------------------
    #  output the result
    #---------------------------------------------------------------------------
    my $outFileName = "$s.domain";

    open( WR, ">$dataset_path/$outFileName" );
    printf WR "$s\t$type\t$M0\n";
    printf WR "@arrayDomainSearch_good\n";
    printf WR "@arrayDPSearch_good\n";
    printf WR "@arrayDomainSearch_bad\n";
    printf WR "@arrayDPSearch_bad\n";
    printf WR "@arrayGapSearch\n";
    printf WR "@arrayTerminalSearch\n";
    close(WR);
    open( WR, ">$dataset_path/$s.tpl" );

    foreach my $line (@arrayTemplateInfo) {
        print WR "$line\n";
    }
    close(WR);
    return;
## Please see file perltidy.ERR
}    # ----------  end of subroutine  match ----------

sub initialize {
    my ($par1) = @_;
    @inits = qw(
      PPA
      dPPA
      dPPA2
      sPPA
      MUSTER
      wPPA
      wdPPA
      wMUSTER
    );

    $M       = 11;    #### in an order of confidence
    %zscore0 = (
        'MUSTER'  => 5.8,
        'dPPA'    => 9.3,
        'wdPPA'   => 9.3,
        'wPPA'    => 7.0,
        'wMUSTER' => 5.8,
        'dPPA2'   => 10.5,
        'PPA'     => 7.0,
        'sPPA'    => 8.0,
    );

    %method2Map = (
        'MUSTER'  => "PDB",
        'dPPA'    => "PDB",
        'wdPPA'   => "PDB",
        'wPPA'    => "PDB",
        'wMUSTER' => "PDB",
        'dPPA2'   => "PDB",
        'PPA'     => "PDB",
        'sPPA'    => "PDB",

    );

    return;
}    # ----------  end of subroutine main  ----------

sub normalizeTemplateName {
    my ($templateName) = @_;
    my $value = "";
    if ( $templateName =~ /:/ ) {
        my @array = split( /:/, $templateName );
        $templateName = $array[1];
        if ( $templateName =~ /^hs/ ) {
            $templateName = substr( $templateName, 2 );
        }
    }

    #    if ( ( $templateName =~ m/[0-9]$/ ) == 0 ) {
    $templateName =~ s/_//g;
    $templateName =~ tr/[A-Z]/[a-z]/;
    my $str1 = substr( $templateName, 0, 4 );
    my $str2 = substr( $templateName, 4, 1 );
    if ( $str2 eq "" ) {
        $str2 = "A";
    }
    if ( $str2 eq "1" ) {
        $str2 = "A";
    }
    if ( $str2 eq "2" ) {
        $str2 = "A";
    }
    if ( $str2 eq "3" ) {
        $str2 = "A";
    }
    $value = $str1 . uc($str2);

    return $value;
}    # ----------  end of subroutine normalizeTemplateName  ----------

sub getMyCathDomainInfo {
    my ($s) = @_;
    my $record = $myCathDB{$s};
    my %data;
    if ( not $record ) {

        #	  print " No information of $s in cath DB.\n ";
    }
    else {
        $record =~ /(\S+)\t(\S+)\t(\S+)\t(\S+)/;
        my $chainName  = $1;
        my $numDomains = $3;
        my $detail     = $4;
        $data{"chainName"}  = $chainName;
        $data{"numDomains"} = $numDomains;

        #	$data{"numFragments"} = $numFragments;
        $detail =~ s/\(//g;
        chop($detail);
        my @array = split( '\)', $detail );
        for ( my $i = 1 ; $i <= $numDomains ; $i++ ) {
            $data{"domain$i"} = $array[ $i - 1 ];
        }

    }

    return %data;
}    # ----------  end of subroutine getCathDomainInfo  ----------

sub getCathDomainInfo {
    my ($s) = @_;
    my %domainInfo;
    my $record = &getRecord($s);

    if ( !$record ) {

        #	    print " No information of $s in cath DB.\n ";

    }
    else {
        %domainInfo = &parseCathRecord($record);
    }

    return %domainInfo;
}    # ----------  end of subroutine getCathDomainInfo  ----------

sub parseCathRecord {
    my ($record) = @_;
    my %data;

    $record =~ /(\S+)\s(\S+)\s(\S+)\s+(.*)/;

    my $chainName    = $1;    #chainName
    my $numDomains   = $2;    #number of domain "D02"
    my $numFragments = $3;    #number of Fragment "F01"
    my $detail       = $4;    #other information

    $numDomains =~ /\S([1-9]+)/;    # "D01" ->1
    $numDomains = $1;
    $numFragments =~ /\S([1-9]+)/;    #  "F01"->1
    $numFragments         = $1;
    $data{"chainName"}    = $chainName;
    $data{"numDomains"}   = $numDomains;
    $data{"numFragments"} = $numFragments;

    for ( my $i = 1 ; $i <= $numDomains ; $i++ ) {
        my $scale = "";

        $detail =~ /\s*(\d+)\s(.*)/;
        my $num = $1;
        $detail = $2;

        for ( my $j = 0 ; $j < $num ; $j++ ) {

            $detail =~ /\s*\S\s+(\S+)\s\S\s\S\s+(\S+)\s\S(.*)/;
            if ( $scale eq "" ) {
                $scale = "$1-$2";
            }
            else {
                $scale .= "|$1-$2";
            }
            $detail = $3;
        }
        $data{"domain$i"} = $scale;

    }

    return %data;
}    # ----------  end of subroutine parseCathRecord  ----------

sub initializeCathDB {
    my ($dbFile) = @_;
    open( my $readcath, "<$dbFile" );
    my $line = "";
    while ( $line = <$readcath> ) {
        if ( !( $line =~ m/^#/ ) ) {
            my $chainName = substr( $line, 0, 5 );
            $cathDB{$chainName} = $line;
        }
    }
    close($readcath);
}    # ----------  end of subroutine initalizeDB  ----------

sub readMyCathDB {
    my ($dbFile) = @_;
    open( my $readcath, "<$dbFile" );
    my $line = "";
    while ( $line = <$readcath> ) {
        chop($line);
        my $chainName = substr( $line, 0, 5 );
        $myCathDB{$chainName} = $line;
    }
    close($readcath);
}    # ----------  end of subroutine initalizeDB  ----------

#===  FUNCTION  ================================================================
#         NAME:  getRecord
#      PURPOSE:
#   PARAMETERS:  ????
#      RETURNS:  ????
#  DESCRIPTION:  ????
#       THROWS:  no exceptions
#     COMMENTS:  none
#     SEE ALSO:  n/a
#===============================================================================
sub getRecord {
    my ($chainName) = @_;
    my $record;
    $record = $cathDB{$chainName};
    return $record;
}    # ----------  end of subroutine getRecord  ----------

#===  FUNCTION  ================================================================
#         NAME:  getRecord
#      PURPOSE:  search a value from an array,and return the index.
#   PARAMETERS:  $arrayRef, a reference of searched Array
#                $value, the value to be search in the array.
#      RETURNS:  $index
#  DESCRIPTION:
#       THROWS:  no exceptions
#     COMMENTS:  none
#     SEE ALSO:  n/a
#===============================================================================

sub searchArrayIndex {
    my ( $arrayRef, $value ) = @_;
    my $nIndex_N = 0;
    my $nIndex_C = $#$arrayRef;
    my $nIndex   = 0;
    for ( my $i = 0 ; $i <= $#$arrayRef ; $i++ ) {
        if ( $$arrayRef[$i] <= $value ) {
            $nIndex_N = $i;
        }
        else {
            last;
        }
    }
    if ( $$arrayRef[$nIndex_N] != $value ) {
        for ( my $i = $#$arrayRef ; $i >= 0 ; $i-- ) {
            if ( $$arrayRef[$i] >= $value ) {
                $nIndex_C = $i;
            }
            else {
                last;
            }
        }
        $nIndex =
          abs( $value - $$arrayRef[$nIndex_N] ) <
          abs( $value - $$arrayRef[$nIndex_C] ) ? $nIndex_N : $nIndex_C;
    }
    else {
        $nIndex = $nIndex_N;
    }

    $nIndex = $nIndex_N;    # set to test.
    return $nIndex;
}    # ----------  end of subroutine searchArrayIndex  ----------

sub searchMapIndex {
    my ( $mapType, $chainID, $pos ) = @_;
    my $aMapRef = $mapDB{$mapType};
    if ( $chainID =~ /:/ ) {
        my @temp = split( /:/, $chainID );
        $chainID = $temp[1];
        if ( $chainID =~ /^hs/ ) {
            $chainID = substr( $chainID, 2 );
        }

    }
    my $length = $$aMapRef{ $chainID, 'length' };
    my $maps   = $$aMapRef{ $chainID, 'maps' };
    if ( not defined $pos ) {
#        print("==$chainID==\n");
    }
    my $index = $pos;
    if ( !$maps ) {
#        print "$chainID no exists in mapDB\n";
        return $index;
    }
    my @array = split( /\t/, $maps );
    my $num = 0;
    my @temp;
    for ( my $i = 0 ; $i <= $#array ; $i++ ) {
        my $str = $array[$i];
        $str =~ /(\d+)-(\d+)/;
        for ( my $j = $1 ; $j <= $2 ; $j++ ) {
            $temp[$num] = $j;
            $num++;
        }

    }

    if ( $temp[0] - $pos > 20 || $pos - $temp[ $num - 1 ] > 20 ) {
        $index = -1;
    }
    elsif ( $temp[0] - $pos > 0 && $temp[0] - $pos <= 20 ) {
        $index = 1;
    }
    elsif ( $pos - $temp[ $num - 1 ] > 0 && $pos - $temp[ $num - 1 ] <= 20 ) {
        $index = $num;
    }
    else {
        for ( my $i = 0 ; $i < $num ; $i++ ) {
            if ( $pos >= $temp[$i] ) {
                $index = $i;
            }
            else {
                last;
            }
        }
    }

    return $index;
}

sub getDPDomainInfo {
    my ($s)       = @_;
    my $aTemplate = &normalizeTemplateName($s);
#    print "================$s=================\n";
    my $record    = $dpDB{"$aTemplate"};
    if ( !$record ) {
        my @domainInfo = &callDomainParser($aTemplate);
        if ( $#domainInfo > 0 ) {
            $dpDB{"$aTemplate"} = \@domainInfo;
            $record = \@domainInfo;
        }
        else {
            $GnumTemplates++;
        }

    }
    return $record;
}

sub callDomainParser {
    my ($s) = @_;
    my $str = "";
    my @boundaries;
    my $index   = 0;
    my $pdbFile = "$pdball_path/$s.pdb";
    if ( !-e "$pdbFile" ) {
        print "$pdbFile not exist\n";
        return @boundaries;
    }
    `/bin/cp "$pdbFile" ./tmp.pdb`;
    my $domain = `"$domainparser_dir/domainparser2" tmp`;
    `rm tmp.pdb`;
    if ( $domain ne "" ) {
        $domain =~ /\S+\s+(\S+)\s+(\d+)/;
        my $Lch      = $1;
        my $n_domain = $2;

        #        printf "Lch=$Lch   n_domain=$n_domain\n";

        if ( $n_domain == 1 ) {
            $boundaries[0] = "1";
            $boundaries[1] = "$Lch";
        }
        else {
            if ( $domain =~ /\S+\s+\S+\s+(\d+)\s+(.+)/ ) {
                my $DOM      = $2;
                my @doms     = split ' ', $DOM;
                my $i_domain = 0;
                foreach my $dom (@doms) {
                    $i_domain++;
                    my @ijs = split ';', $dom;
                    foreach my $ij (@ijs) {
                        $ij =~ /(\d+)-(\d+)/;
                        $boundaries[$index] = $1;
                        $index++;
                        $boundaries[$index] = $2;
                        $index++;
                    }
                }
            }
        }
        @boundaries = sort { $a <=> $b } @boundaries;

        #        print "$s:Domainparser:@boundaries\n";
    }
    else {
        print "DomainPaser Failed!";
    }
    return @boundaries;
}

sub initializeMapDB {
    my ($mapDir) = @_;

#-------------------------------------------------------------------------------
# Threading library mapping. If there multi-library for different threading.
# ThreaDom will mapping them to a unified format,PDB files. The mapping data was
# put in the directory ../map. for example:
#      my @mapTypes = qw(HHM HOMSTRAD PDB SP3 XML ), there is 5 library formats.
#-------------------------------------------------------------------------------
#my @mapTypes = qw(HHM HOMSTRAD PDB SP3 XML );

 my @mapTypes = qw(PDB); #only one format used in this version, other threading was developed by other groups.
    foreach my $mapType (@mapTypes) {
        my $mapFile = "$mapDir/$mapType.map";
        #print "$mapFile---------\n";
        my $reader;
        my %aMap;
        my $line = "";
        if ( !-e "$mapFile" ) {
            print "$mapFile is not exist.";
            exit();
        }
        open( $reader, "<$mapFile" );
        <$reader>;
        while ( $line = <$reader> ) {
            $line =~ /(\S+)\t(\d+)\t(.*)/;
            my $chainid     = $1;
            my $chainLength = $2;
            my $chainMap    = $3;
            if ($3) {
                $aMap{ $chainid, "maps" }   = $chainMap;
                $aMap{ $chainid, 'length' } = $chainLength;
            }
        }
        close($reader);
        $mapDB{$mapType} = \%aMap;
    }
    return %mapDB;
}

sub writeDpDB {
    my ($outFile) = @_;
    my $wr;
    open( $wr, ">$outFile" );
    while ( my ( $key, $value ) = each %dpDB ) {
        printf $wr "$key\t@$value\n";
    }
    close($wr);
}

sub readDpDB {
    my ($inFile) = @_;

    my $line = "";
    if ( -e "$inFile" ) {
        open( RD, "<$inFile" );
        while ( $line = <RD> ) {
            $line =~ /(\S+)\t(.*)/;
            my $template = $1;

            my $links = $2;
            my @array = split( /\t/, $links );
            $dpDB{$template} = \@array;
        }

        close(RD);
    }
    else {
        print "$inFile not exist.";
    }
}

sub predict {
    my (
        $datasetBase, $datasetName, $chainId,
        $cutoff_easy, $cutoff_medm, $cutoff_hard
    ) = @_;
    my $mtadom = new ThreaDom(
        BaseName    => $datasetBase,
        ChainId     => $chainId,
        DatasetName => $datasetName
    );
    my $type = $mtadom->{ChainAttributes}->{Type};
    my $cutoff;
    if ( $type eq 'easy' ) {
        $cutoff = $cutoff_easy;

    }
    elsif ( $type eq 'medm' ) {
        $cutoff = $cutoff_medm;
    }
    elsif ( $type eq 'hard' ) {
        $cutoff = $cutoff_hard;
    }

    #$mtadom->printScore($chainId,"$chainId.score");
    $mtadom->smooth(20);

    $mtadom->printScore( $chainId,
        "$dataset_path/$chainId.score" );
    my $str             = $mtadom->process($cutoff);
    my $straightPredict = $mtadom->getStraightPredcition();
`echo "$straightPredict" > "$dataset_path/$chainId.result"`;
 print "$straightPredict\n";
}

