package ThreaDom;
use strict;
use warnings;
use DomainHelper;

sub new {
	my ( $invocant, $class, $self );
	$invocant = shift;                          # either class name or object
	$class    = ref($invocant) || $invocant;    # object or class name
	$self     = {};
	bless( $self, $class );                     # bless $self into class $class
	$self->_init(@_);
	return $self;
}

sub _init {
	my $self = shift;
	my ( %args, $chainId, $BaseName, $datasetName );
	%args                    = @_;
	$BaseName                = $args{BaseName};
	$chainId                 = $args{ChainId};
	$datasetName             = $args{DatasetName};
	$self->{BaseName}        = $BaseName;
	$self->{ChainId}         = $chainId;
	$self->{ChainAttributes} = ();
	$self->{DatasetName}     = $datasetName;
	$self->{Config}          = ();
	$self->{CathArray_GOOD}  = [];
	$self->{DpArray_BAD}     = [];
	$self->{CathArray_BAD}   = [];
	$self->{DpArray_GOOD}    = [];
	$self->{GapArray}        = [];
	$self->{TerminalArray}   = [];

	# easy: [2.0,0.6,0.8,0.1],50,12
	# medm: [1.5,0.8,1.0,0.4] ,50 ,12;
	# hard:[0,0.8,1.2,0.5] ,50,11;
	$self->{Weight}         = [ 2.0, 0.6, 0.8, 0.1 ];
	$self->{NT}             = 50;
	$self->{Window}         = 12;
	$self->{GinputArray}    = [];
	$self->{GoutputArray}   = [];
	$self->{GboundaryArray} = [];
	$self->{GdiffArray}     = [];
	$self->{Test}           = ();

	#	$self->{SecondarayStructure} = "";
	#	$self->{Exposed}             = [];

	$self->_readMatch();
	$self->_score();

}

sub _setParameter {
	my $self = shift;
	my $type = $self->{ChainAttributes}->{Type};
	if ( $type eq 'easy' ) {

		$self->{Weight} = [ 2.0, 0.6, 0.8, 0.1 ];
		$self->{NT}     = 50;
		$self->{Window} = 12;

	}
	elsif ( $type eq 'medm' ) {

		$self->{Weight} = [ 2.0, 0.5, 1.1, 0.5 ];
		$self->{NT}     = 50;
		$self->{Window} = 12;
	}
	elsif ( $type eq 'hard' ) {

		$self->{Weight} = [ 0.0, 0.5, 1.4, 0.5 ];
		$self->{NT}     = 50;
		$self->{Window} = 10;
	}

}

sub setWeight {
	my $self   = shift;
	my $weight = @_;
	$self->{Weight} = $weight;
}

sub setPenaltyWindow {
	my $self   = shift;
	my $window = @_;
	$self->{Window} = $window;
}

sub setTemplateNum {
	my $self = shift;
	my $num  = @_;
	$self->{NT} = $num;
}

sub smooth {
	my $self       = shift;
	my ($window)   = @_;
	my $n          = $window / 2;
	my $nLastIndex = $#{ $self->{GinputArray} };
	my @tempArray;

	for ( my $i = 0 ; $i <= $nLastIndex ; $i++ ) {
		my $sum = $self->{GinputArray}->[$i];
		for ( my $j = 1 ; $j < $n ; $j++ ) {
			if ( ( $i + $j ) <= $nLastIndex ) {
				$sum += $self->{GinputArray}->[ $i + $j ];
			}
			if ( ( $i - $j >= 0 ) ) {
				$sum += $self->{GinputArray}->[ $i - $j ];
			}
		}
		$tempArray[$i] = $sum / ( $window - 1 );
	}
	$self->{GinputArray} = \@tempArray;
}

sub process {
	my $self       = shift;
	my ($cutoff)   = @_;
	my $nLastIndex = $#{ $self->{GinputArray} };
	my $Lch        = $nLastIndex + 1;

	for ( my $i = 0 ; $i <= $nLastIndex ; $i++ ) {

		if ( $self->{GinputArray}->[$i] < $cutoff ) {
			$self->{GoutputArray}->[$i] = 0;
		}
		else {
			$self->{GoutputArray}->[$i] = 1;
		}
	}

	for ( my $i = 0 ; $i <= $nLastIndex ; $i++ ) {
		$self->{GboundaryArray}->[$i] =
		  $self->{GinputArray}->[$i] * $self->{GoutputArray}->[$i];
	}

	$self->{GdiffArray}->[0] = -1;
	for ( my $i = 1 ; $i < $nLastIndex ; $i++ ) {
		$self->{GdiffArray}->[$i] =
		  $self->{GoutputArray}->[$i] - $self->{GoutputArray}->[ $i - 1 ];
	}
	$self->{GdiffArray}->[$nLastIndex] = 1;

	my $nStart   = 1;
	my $nEnd     = $Lch;
	my $str      = "";
	my $linksNum = 0;
	for ( my $i = 10 ; $i < $nLastIndex - 10 ; $i++ ) {
		if ( $self->{GdiffArray}->[$i] == -1 ) {
			$nStart = $i;
			$nStart = 0 if ( $nStart < 20 );
			$nEnd   = $Lch;
		}
		if ( $self->{GdiffArray}->[$i] == 1 ) {
			$nEnd = $i - 1;
			$nEnd = $Lch if ( $nEnd > $Lch - 20 );
			if ( !( $nEnd < 66 ) && !( $nStart > $Lch - 66 ) ) {
				my $average = ( $nStart + $nEnd ) / 2;
				if ( ( $average > 40 ) && ( $average < $Lch - 40 ) ) {

					$str .= "($nStart-$nEnd)";
					$linksNum++;
				}
			}
			$nStart = 0;
		}
	}

	$self->{LinksNum}       = $linksNum;
	$self->{PredictedLinks} = $str;
	return $str;
}

sub getStraightPredcition {
	my $self = shift;
	if ( not defined $self->{PredictedLinks} ) {
		print("no links predicion result.call process(\$cutoff) first.\n");
		return "";
	}
	my $chainId        = $self->{ChainId};
	my $length         = $self->{ChainAttributes}->{Length};
	my $predictedLinks = $self->{PredictedLinks};
	my $helper         = new DomainHelper();
	my $domains        = $helper->links2Domains( $predictedLinks, $length );
	my $predictedNums  = $self->{LinksNum} + 1;
	my $str            = "$chainId\t$length\t$predictedNums\t$domains";
	return $str;
}

sub _penalty {
	my $self = shift;
	my ( $nPos, $penalty ) = @_;
	my $arrayDomainSearch = $self->{GinputArray};
	my $length            = $#$arrayDomainSearch + 1;
	if ( $nPos > 0 && $nPos < $length ) {
		$$arrayDomainSearch[$nPos] -= $penalty;
		for ( my $n = 1 ; $n < $self->{Window} ; $n++ ) {
			if ( $nPos - $n >= 0 ) {
				$$arrayDomainSearch[ $nPos - $n ] -= $penalty;
			}
			if ( $nPos + $n < $length ) {
				$$arrayDomainSearch[ $nPos + $n ] -= $penalty;
			}
		}
	}
}

sub _score {
	my $self = shift;
	my ( $w1, $w2, $w3, $w4 );

	$self->_setParameter();    # set weight,window and template nums.

	$w1 = $self->{Weight}->[0];
	$w2 = $self->{Weight}->[1];
	$w3 = $self->{Weight}->[2];
	$w4 = $self->{Weight}->[3];
	my $cathArray_good = $self->{CathArray_GOOD};
	my $dpArray_good   = $self->{DpArray_GOOD};
	my $cathArray_bad  = $self->{CathArray_BAD};
	my $dpArray_bad    = $self->{DpArray_BAD};

	my $gapArray      = $self->{GapArray};
	my $terminalArray = $self->{TerminalArray};
	for ( my $i = 0 ; $i <= $#$cathArray_good ; $i++ ) {
		$self->{GinputArray}->[$i] = $self->{NT};
	}

	for ( my $i = 0 ; $i <= $#$cathArray_good ; $i++ ) {
		if ( $cathArray_good->[$i] > 0 ) {
			$self->_penalty( $i, $w1 * $cathArray_good->[$i] );
		}

		if ( $dpArray_good->[$i] > 0 ) {
			$self->_penalty( $i, 0.8 * $w1 * $dpArray_good->[$i] );
		}
		if ( $cathArray_bad->[$i] > 0 ) {
			$self->_penalty( $i, $w2 * $cathArray_bad->[$i] );
		}

		if ( $dpArray_bad->[$i] > 0 ) {
			$self->_penalty( $i, 0.8 * $w2 * $dpArray_bad->[$i] );
		}

		if ( $terminalArray->[$i] > 0 ) {
			$self->_penalty( $i, $w3 * $terminalArray->[$i] );
		}
		if ( $gapArray->[$i] > 0 ) {
			$self->_penalty( $i, $w4 * $gapArray->[$i] );
		}

	}

}

sub printScore {
	my $self = shift;
	my ( $chainId, $outFile ) = @_;
        if ( $self->{ChainId} eq $chainId ) {
		open( WR, ">$outFile" );
                 my $str='';
                foreach my $var( @{$self->{GinputArray}}){
                 $var=$var/50.0;
		 $str.="$var ";
                }
                chop($str);
                printf WR "$str\n";
		close(WR);
	}

}

sub _readMatch {

	my $self        = shift;
	my $datasetName = $self->{DatasetName};
	my $chainId=$self->{ChainId};
#	my $inputFileName =
#	  "$self->{BaseName}/$datasetName/$chainId/$chainId.domain";
	
	my $inputFileName =
	  "$self->{BaseName}/$chainId.domain";
    open my $r_input, "$inputFileName" or die "can't open $inputFileName";
	my $line;
	$line = <$r_input>;    #we can save more information in this line.
	$line =~ /(\S+)\t(\S+)\t(\d+)\n/;
	my $targetName = $1;

	# Here we can save more shared attribures of the chains.
	$self->{ChainAttributes}->{Type} = $2;
	$self->{ChainAttributes}->{M0}   = $3;
	$line                            = <$r_input>;
	my @cathArray_good = split( /\s/, $line );
	$self->{ChainAttributes}->{Length} = $#cathArray_good + 1;

	######################################################
	$line = <$r_input>;
	my @dpArray_good = split( /\s/, $line );

	$line = <$r_input>;
	my @cathArray_bad = split( /\s/, $line );

	$line = <$r_input>;
	my @dpArray_bad = split( /\s/, $line );
	$line = <$r_input>;
	my @gapArray = split( /\s/, $line );
	$line = <$r_input>;
	my @terminalArray = split( /\s/, $line );

	$self->{CathArray_GOOD} = \@cathArray_good;
	$self->{DpArray_GOOD}   = \@dpArray_good;
	$self->{CathArray_BAD}  = \@cathArray_bad;
	$self->{DpArray_BAD}    = \@dpArray_bad;
	$self->{GapArray}       = \@gapArray;
	$self->{TerminalArray}  = \@terminalArray;
	close($r_input);

}

1;
